#/usr/bin/env bash

duvet extract https://www.rfc-editor.org/rfc/rfc5246 # The Transport Layer Security (TLS) Protocol Version 1.2
duvet extract https://www.rfc-editor.org/rfc/rfc5869 # HMAC-based Extract-and-Expand Key Derivation Function (HKDF)
duvet extract https://www.rfc-editor.org/rfc/rfc8446 # The Transport Layer Security (TLS) Protocol Version 1.3
duvet extract https://www.rfc-editor.org/rfc/rfc8448 # Example Handshake Traces for TLS 1.3
duvet extract https://www.rfc-editor.org/rfc/rfc7627 # Transport Layer Security (TLS) Session Hash and Extended Master Secret Extension
duvet extract https://www.rfc-editor.org/rfc/rfc5746 # Transport Layer Security (TLS) Renegotiation Indication Extension
duvet extract https://www.rfc-editor.org/rfc/rfc4492 # Elliptic Curve Cryptography (ECC) Cipher Suites for Transport Layer Security (TLS)
duvet extract https://www.rfc-editor.org/rfc/rfc8422 # Elliptic Curve Cryptography (ECC) Cipher Suites for Transport Layer Security (TLS) Versions 1.2 and Earlier
